import ctypes
from typing import Self

import numpy as np
import polars as pl
from pyautd3.native_methods.utils import _validate_ptr

from pyautd3_emulator.instant import Instant, InstantRecordOption
from pyautd3_emulator.native_methods.autd3capi_emulator import NativeMethods as Emu
from pyautd3_emulator.native_methods.autd3capi_emulator import RecordPtr
from pyautd3_emulator.range import RangeXYZ
from pyautd3_emulator.rms import Rms, RmsRecordOption


class Record:
    _ptr: RecordPtr

    def __init__(self: Self, ptr: RecordPtr) -> None:
        self._ptr = ptr

    def __del__(self: Self) -> None:
        self._dispose()

    def _dispose(self: Self) -> None:
        if self._ptr.value is not None:  # pragma: no cover
            Emu().emulator_record_free(self._ptr)
            self._ptr.value = None

    def phase(self: Self) -> pl.DataFrame:
        cols = int(Emu().emulator_record_drive_cols(self._ptr))
        rows = int(Emu().emulator_record_drive_rows(self._ptr))
        time = np.zeros(cols, dtype=np.uint64)
        v = np.zeros([cols, rows], dtype=np.uint8)
        Emu().emulator_record_phase(
            self._ptr,
            time.ctypes.data_as(ctypes.POINTER(ctypes.c_uint64)),  # type: ignore[arg-type]
            ctypes.cast(
                (ctypes.POINTER(ctypes.c_uint8) * cols)(
                    *(ctypes.cast(r, ctypes.POINTER(ctypes.c_uint8)) for r in np.ctypeslib.as_ctypes(v)),  # type: ignore[arg-type]
                ),
                ctypes.POINTER(ctypes.POINTER(ctypes.c_uint8)),
            ),
        )
        return pl.DataFrame({s.name: s for s in (pl.Series(name=f"phase@{time[i]}[ns]", values=v[i]) for i in range(cols))})

    def pulse_width(self: Self) -> pl.DataFrame:
        cols = int(Emu().emulator_record_drive_cols(self._ptr))
        rows = int(Emu().emulator_record_drive_rows(self._ptr))
        time = np.zeros(cols, dtype=np.uint64)
        v = np.zeros([cols, rows], dtype=np.uint8)
        Emu().emulator_record_pulse_width(
            self._ptr,
            time.ctypes.data_as(ctypes.POINTER(ctypes.c_uint64)),  # type: ignore[arg-type]
            ctypes.cast(
                (ctypes.POINTER(ctypes.c_uint8) * cols)(
                    *(ctypes.cast(r, ctypes.POINTER(ctypes.c_uint8)) for r in np.ctypeslib.as_ctypes(v)),  # type: ignore[arg-type]
                ),
                ctypes.POINTER(ctypes.POINTER(ctypes.c_uint8)),
            ),
        )
        return pl.DataFrame({s.name: s for s in (pl.Series(name=f"pulse_width@{time[i]}[ns]", values=v[i]) for i in range(cols))})

    def output_voltage(self: Self) -> pl.DataFrame:
        cols = int(Emu().emulator_record_output_cols(self._ptr))
        rows = int(Emu().emulator_record_drive_rows(self._ptr))
        v = np.zeros([cols, rows], dtype=np.float32)
        Emu().emulator_record_output_voltage(
            self._ptr,
            ctypes.cast(
                (ctypes.POINTER(ctypes.c_float) * cols)(
                    *(ctypes.cast(r, ctypes.POINTER(ctypes.c_float)) for r in np.ctypeslib.as_ctypes(v)),  # type: ignore[arg-type]
                ),
                ctypes.POINTER(ctypes.POINTER(ctypes.c_float)),
            ),
        )
        return pl.DataFrame({s.name: s for s in (pl.Series(name=f"voltage[V]@{i}[25us/256]", values=v[i]) for i in range(cols))})

    def output_ultrasound(self: Self) -> pl.DataFrame:
        cols = int(Emu().emulator_record_output_cols(self._ptr))
        rows = int(Emu().emulator_record_drive_rows(self._ptr))
        v = np.zeros([cols, rows], dtype=np.float32)
        Emu().emulator_record_output_ultrasound(
            self._ptr,
            ctypes.cast(
                (ctypes.POINTER(ctypes.c_float) * cols)(
                    *(ctypes.cast(r, ctypes.POINTER(ctypes.c_float)) for r in np.ctypeslib.as_ctypes(v)),  # type: ignore[arg-type]
                ),
                ctypes.POINTER(ctypes.POINTER(ctypes.c_float)),
            ),
        )
        return pl.DataFrame({s.name: s for s in (pl.Series(name=f"p[a.u.]@{i}[25us/256]", values=v[i]) for i in range(cols))})

    def sound_field(self: Self, range_: RangeXYZ, option: InstantRecordOption | RmsRecordOption) -> Instant | Rms:
        match option:
            case InstantRecordOption():
                return Instant(
                    _validate_ptr(
                        Emu().emulator_sound_field_instant(self._ptr, range_._inner, option._inner),
                    ),
                )
            case RmsRecordOption():  # pragma: no cover
                return Rms(
                    _validate_ptr(
                        Emu().emulator_sound_field_rms(self._ptr, range_._inner, option._inner),
                    ),
                )



import ctypes
from collections.abc import Callable, Iterable
from types import TracebackType
from typing import Self

import numpy as np
import polars as pl
from pyautd3.driver.autd3_device import AUTD3
from pyautd3.driver.geometry.geometry import Geometry
from pyautd3.ethercat.dc_sys_time import DcSysTime
from pyautd3.native_methods.autd3 import DcSysTime as _DcSysTime
from pyautd3.native_methods.autd3capi import ControllerPtr
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.structs import Point3, Quaternion
from pyautd3.native_methods.utils import _validate_ptr

from pyautd3_emulator.native_methods.autd3capi_emulator import EmulatorPtr
from pyautd3_emulator.native_methods.autd3capi_emulator import NativeMethods as Emu
from pyautd3_emulator.record import Record
from pyautd3_emulator.recorder import Recorder


class Emulator(Geometry):
    _ptr: EmulatorPtr

    def __init__(self: Self, devices: Iterable[AUTD3]) -> None:
        devices = list(devices)
        pos = np.fromiter((np.void(Point3(d.pos)) for d in devices), dtype=Point3)  # type: ignore[type-var,call-overload]
        rot = np.fromiter((np.void(Quaternion(d.rot)) for d in devices), dtype=Quaternion)  # type: ignore[type-var,call-overload]
        self._ptr = Emu().emulator(
            pos.ctypes.data_as(ctypes.POINTER(Point3)),  # type: ignore[arg-type]
            rot.ctypes.data_as(ctypes.POINTER(Quaternion)),  # type: ignore[arg-type]
            len(devices),
        )
        super().__init__(Emu().emulator_geometry(self._ptr))

    @property
    def geometry(self: Self) -> Geometry:
        return self

    def transducer_table(self: Self) -> pl.DataFrame:
        n = int(Emu().emulator_transducer_table_rows(self._ptr))
        dev_indices = np.zeros(n, dtype=np.uint16)
        tr_indices = np.zeros(n, dtype=np.uint8)
        x = np.zeros(n, dtype=np.float32)
        y = np.zeros(n, dtype=np.float32)
        z = np.zeros(n, dtype=np.float32)
        nx = np.zeros(n, dtype=np.float32)
        ny = np.zeros(n, dtype=np.float32)
        nz = np.zeros(n, dtype=np.float32)
        Emu().emulator_transducer_table(
            self._ptr,
            dev_indices.ctypes.data_as(ctypes.POINTER(ctypes.c_uint16)),  # type: ignore[arg-type]
            tr_indices.ctypes.data_as(ctypes.POINTER(ctypes.c_uint8)),  # type: ignore[arg-type]
            x.ctypes.data_as(ctypes.POINTER(ctypes.c_float)),  # type: ignore[arg-type]
            y.ctypes.data_as(ctypes.POINTER(ctypes.c_float)),  # type: ignore[arg-type]
            z.ctypes.data_as(ctypes.POINTER(ctypes.c_float)),  # type: ignore[arg-type]
            nx.ctypes.data_as(ctypes.POINTER(ctypes.c_float)),  # type: ignore[arg-type]
            ny.ctypes.data_as(ctypes.POINTER(ctypes.c_float)),  # type: ignore[arg-type]
            nz.ctypes.data_as(ctypes.POINTER(ctypes.c_float)),  # type: ignore[arg-type]
        )
        return pl.DataFrame(
            {
                "dev_idx": dev_indices,
                "tr_idx": tr_indices,
                "x[mm]": x,
                "y[mm]": y,
                "z[mm]": z,
                "nx": nx,
                "ny": ny,
                "nz": nz,
            },
        )

    def record(self: Self, f: Callable[[Recorder], None]) -> Record:
        return self.record_from(DcSysTime.__private_new__(_DcSysTime(0)), f)

    def record_from(self: Self, start_time: DcSysTime, f: Callable[[Recorder], None]) -> Record:
        def f_native(ptr: ControllerPtr) -> None:
            geometry = Base().geometry(ptr)
            cnt = Recorder(geometry, ptr)
            f(cnt)
            cnt._disposed = True

        f_native_ = ctypes.CFUNCTYPE(None, ControllerPtr)(f_native)

        return Record(
            _validate_ptr(
                Emu().emulator_record_from(self._ptr, start_time._inner, f_native_),  # type: ignore[arg-type]
            ),
        )

    def __del__(self: Self) -> None:
        self._dispose()

    def _dispose(self: Self) -> None:
        if self._ptr.value is not None:
            Emu().emulator_free(self._ptr)
            self._ptr.value = None

    def __enter__(self: Self) -> Self:
        return self

    def __exit__(
        self: Self,
        _exc_type: type[BaseException] | None,
        _exc_value: BaseException | None,
        _traceback: TracebackType | None,
    ) -> None:
        self._dispose()



import ctypes
import threading
from pathlib import Path

from pyautd3.native_methods.autd3 import DcSysTime
from pyautd3.native_methods.autd3capi_driver import Duration, GeometryPtr, LinkPtr, ResultStatus
from pyautd3.native_methods.structs import Point3, Quaternion


class RecordPtr(ctypes.Structure):
    _fields_ = [("value", ctypes.c_void_p)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, RecordPtr) and self._fields_ == other._fields_  # pragma: no cover


class RmsPtr(ctypes.Structure):
    _fields_ = [("value", ctypes.c_void_p)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, RmsPtr) and self._fields_ == other._fields_  # pragma: no cover


class ResultRecord(ctypes.Structure):
    _fields_ = [("result", RecordPtr), ("err_len", ctypes.c_uint32), ("err", ctypes.c_void_p)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, ResultRecord) and self._fields_ == other._fields_  # pragma: no cover


class EmulatorControllerPtr(ctypes.Structure):
    _fields_ = [("value", ctypes.c_void_p)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, EmulatorControllerPtr) and self._fields_ == other._fields_  # pragma: no cover


class InstantPtr(ctypes.Structure):
    _fields_ = [("value", ctypes.c_void_p)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, InstantPtr) and self._fields_ == other._fields_  # pragma: no cover


class RangeXYZ(ctypes.Structure):
    _fields_ = [
        ("x_start", ctypes.c_float),
        ("x_end", ctypes.c_float),
        ("y_start", ctypes.c_float),
        ("y_end", ctypes.c_float),
        ("z_start", ctypes.c_float),
        ("z_end", ctypes.c_float),
        ("resolution", ctypes.c_float),
    ]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, RangeXYZ) and self._fields_ == other._fields_  # pragma: no cover


class ResultInstant(ctypes.Structure):
    _fields_ = [("result", InstantPtr), ("err_len", ctypes.c_uint32), ("err", ctypes.c_void_p)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, ResultInstant) and self._fields_ == other._fields_  # pragma: no cover


class RmsRecordOption(ctypes.Structure):
    _fields_ = [("sound_speed", ctypes.c_float), ("print_progress", ctypes.c_bool), ("gpu", ctypes.c_bool)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, RmsRecordOption) and self._fields_ == other._fields_  # pragma: no cover


class EmulatorPtr(ctypes.Structure):
    _fields_ = [("value", ctypes.c_void_p)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, EmulatorPtr) and self._fields_ == other._fields_  # pragma: no cover


class InstantRecordOption(ctypes.Structure):
    _fields_ = [
        ("sound_speed", ctypes.c_float),
        ("time_step", Duration),
        ("print_progress", ctypes.c_bool),
        ("memory_limits_hint_mb", ctypes.c_uint64),
        ("gpu", ctypes.c_bool),
    ]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, InstantRecordOption) and self._fields_ == other._fields_  # pragma: no cover


class ResultRms(ctypes.Structure):
    _fields_ = [("result", RmsPtr), ("err_len", ctypes.c_uint32), ("err", ctypes.c_void_p)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, ResultRms) and self._fields_ == other._fields_  # pragma: no cover


class Singleton(type):
    _instances = {}  # type: ignore[var-annotated]
    _lock = threading.Lock()

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            with cls._lock:
                if cls not in cls._instances:  # pragma: no cover
                    cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


class NativeMethods(metaclass=Singleton):
    def init_dll(self, bin_location: Path, bin_prefix: str, bin_ext: str) -> None:
        self.dll = ctypes.CDLL(str(bin_location / f"{bin_prefix}autd3capi_emulator{bin_ext}"))

        self.dll.AUTDEmulatorRecordDriveCols.argtypes = [RecordPtr]
        self.dll.AUTDEmulatorRecordDriveCols.restype = ctypes.c_uint64

        self.dll.AUTDEmulatorRecordDriveRows.argtypes = [RecordPtr]
        self.dll.AUTDEmulatorRecordDriveRows.restype = ctypes.c_uint64

        self.dll.AUTDEmulatorRecordPhase.argtypes = [RecordPtr, ctypes.POINTER(ctypes.c_uint64), ctypes.POINTER(ctypes.POINTER(ctypes.c_uint8))]
        self.dll.AUTDEmulatorRecordPhase.restype = None

        self.dll.AUTDEmulatorRecordPulseWidth.argtypes = [RecordPtr, ctypes.POINTER(ctypes.c_uint64), ctypes.POINTER(ctypes.POINTER(ctypes.c_uint8))]
        self.dll.AUTDEmulatorRecordPulseWidth.restype = None

        self.dll.AUTDEmulatorSoundFieldInstant.argtypes = [RecordPtr, RangeXYZ, InstantRecordOption]
        self.dll.AUTDEmulatorSoundFieldInstant.restype = ResultInstant

        self.dll.AUTDEmulatorSoundFieldInstantTimeLen.argtypes = [InstantPtr, Duration]
        self.dll.AUTDEmulatorSoundFieldInstantTimeLen.restype = ctypes.c_uint64

        self.dll.AUTDEmulatorSoundFieldInstantPointsLen.argtypes = [InstantPtr]
        self.dll.AUTDEmulatorSoundFieldInstantPointsLen.restype = ctypes.c_uint64

        self.dll.AUTDEmulatorSoundFieldInstantGetX.argtypes = [InstantPtr, ctypes.POINTER(ctypes.c_float)]
        self.dll.AUTDEmulatorSoundFieldInstantGetX.restype = None

        self.dll.AUTDEmulatorSoundFieldInstantGetY.argtypes = [InstantPtr, ctypes.POINTER(ctypes.c_float)]
        self.dll.AUTDEmulatorSoundFieldInstantGetY.restype = None

        self.dll.AUTDEmulatorSoundFieldInstantGetZ.argtypes = [InstantPtr, ctypes.POINTER(ctypes.c_float)]
        self.dll.AUTDEmulatorSoundFieldInstantGetZ.restype = None

        self.dll.AUTDEmulatorSoundFieldInstantSkip.argtypes = [InstantPtr, Duration]
        self.dll.AUTDEmulatorSoundFieldInstantSkip.restype = ResultStatus

        self.dll.AUTDEmulatorSoundFieldInstantNext.argtypes = [
            InstantPtr,
            Duration,
            ctypes.POINTER(ctypes.c_uint64),
            ctypes.POINTER(ctypes.POINTER(ctypes.c_float)),
        ]
        self.dll.AUTDEmulatorSoundFieldInstantNext.restype = ResultStatus

        self.dll.AUTDEmulatorSoundFieldInstantFree.argtypes = [InstantPtr]
        self.dll.AUTDEmulatorSoundFieldInstantFree.restype = None

        self.dll.AUTDEmulatorTracingInit.argtypes = []
        self.dll.AUTDEmulatorTracingInit.restype = None

        self.dll.AUTDEmulatorTracingInitWithFile.argtypes = [ctypes.c_char_p]
        self.dll.AUTDEmulatorTracingInitWithFile.restype = ResultStatus

        self.dll.AUTDEmulator.argtypes = [ctypes.POINTER(Point3), ctypes.POINTER(Quaternion), ctypes.c_uint16]
        self.dll.AUTDEmulator.restype = EmulatorPtr

        self.dll.AUTDEmulatorFree.argtypes = [EmulatorPtr]
        self.dll.AUTDEmulatorFree.restype = None

        self.dll.AUTDEmulatorGeometry.argtypes = [EmulatorPtr]
        self.dll.AUTDEmulatorGeometry.restype = GeometryPtr

        self.dll.AUTDEmulatorRecordFrom.argtypes = [EmulatorPtr, DcSysTime, ctypes.c_void_p]
        self.dll.AUTDEmulatorRecordFrom.restype = ResultRecord

        self.dll.AUTDEmulatorRecordFree.argtypes = [RecordPtr]
        self.dll.AUTDEmulatorRecordFree.restype = None

        self.dll.AUTDEmulatorTickNs.argtypes = [LinkPtr, Duration]
        self.dll.AUTDEmulatorTickNs.restype = ResultStatus

        self.dll.AUTDEmulatorTransducerTableRows.argtypes = [EmulatorPtr]
        self.dll.AUTDEmulatorTransducerTableRows.restype = ctypes.c_uint64

        self.dll.AUTDEmulatorTransducerTable.argtypes = [
            EmulatorPtr,
            ctypes.POINTER(ctypes.c_uint16),
            ctypes.POINTER(ctypes.c_uint8),
            ctypes.POINTER(ctypes.c_float),
            ctypes.POINTER(ctypes.c_float),
            ctypes.POINTER(ctypes.c_float),
            ctypes.POINTER(ctypes.c_float),
            ctypes.POINTER(ctypes.c_float),
            ctypes.POINTER(ctypes.c_float),
        ]
        self.dll.AUTDEmulatorTransducerTable.restype = None

        self.dll.AUTDEmulatorRecordOutputCols.argtypes = [RecordPtr]
        self.dll.AUTDEmulatorRecordOutputCols.restype = ctypes.c_uint64

        self.dll.AUTDEmulatorRecordOutputVoltage.argtypes = [RecordPtr, ctypes.POINTER(ctypes.POINTER(ctypes.c_float))]
        self.dll.AUTDEmulatorRecordOutputVoltage.restype = None

        self.dll.AUTDEmulatorRecordOutputUltrasound.argtypes = [RecordPtr, ctypes.POINTER(ctypes.POINTER(ctypes.c_float))]
        self.dll.AUTDEmulatorRecordOutputUltrasound.restype = None

        self.dll.AUTDEmulatorSoundFieldRms.argtypes = [RecordPtr, RangeXYZ, RmsRecordOption]
        self.dll.AUTDEmulatorSoundFieldRms.restype = ResultRms

        self.dll.AUTDEmulatorSoundFieldRmsTimeLen.argtypes = [RmsPtr, Duration]
        self.dll.AUTDEmulatorSoundFieldRmsTimeLen.restype = ctypes.c_uint64

        self.dll.AUTDEmulatorSoundFieldRmsPointsLen.argtypes = [RmsPtr]
        self.dll.AUTDEmulatorSoundFieldRmsPointsLen.restype = ctypes.c_uint64

        self.dll.AUTDEmulatorSoundFieldRmsGetX.argtypes = [RmsPtr, ctypes.POINTER(ctypes.c_float)]
        self.dll.AUTDEmulatorSoundFieldRmsGetX.restype = None

        self.dll.AUTDEmulatorSoundFieldRmsGetY.argtypes = [RmsPtr, ctypes.POINTER(ctypes.c_float)]
        self.dll.AUTDEmulatorSoundFieldRmsGetY.restype = None

        self.dll.AUTDEmulatorSoundFieldRmsGetZ.argtypes = [RmsPtr, ctypes.POINTER(ctypes.c_float)]
        self.dll.AUTDEmulatorSoundFieldRmsGetZ.restype = None

        self.dll.AUTDEmulatorSoundFieldRmsSkip.argtypes = [RmsPtr, Duration]
        self.dll.AUTDEmulatorSoundFieldRmsSkip.restype = ResultStatus

        self.dll.AUTDEmulatorSoundFieldRmsNext.argtypes = [
            RmsPtr,
            Duration,
            ctypes.POINTER(ctypes.c_uint64),
            ctypes.POINTER(ctypes.POINTER(ctypes.c_float)),
        ]
        self.dll.AUTDEmulatorSoundFieldRmsNext.restype = ResultStatus

        self.dll.AUTDEmulatorSoundFieldRmsFree.argtypes = [RmsPtr]
        self.dll.AUTDEmulatorSoundFieldRmsFree.restype = None

    def emulator_record_drive_cols(self, record: RecordPtr) -> ctypes.c_uint64:
        return self.dll.AUTDEmulatorRecordDriveCols(record)

    def emulator_record_drive_rows(self, record: RecordPtr) -> ctypes.c_uint64:
        return self.dll.AUTDEmulatorRecordDriveRows(record)

    def emulator_record_phase(self, record: RecordPtr, time: ctypes.Array[ctypes.c_uint64], v: ctypes.Array[ctypes.Array[ctypes.c_uint8]]) -> None:
        return self.dll.AUTDEmulatorRecordPhase(record, time, v)

    def emulator_record_pulse_width(
        self,
        record: RecordPtr,
        time: ctypes.Array[ctypes.c_uint64],
        v: ctypes.Array[ctypes.Array[ctypes.c_uint8]],
    ) -> None:
        return self.dll.AUTDEmulatorRecordPulseWidth(record, time, v)

    def emulator_sound_field_instant(self, record: RecordPtr, range_: RangeXYZ, option: InstantRecordOption) -> ResultInstant:
        return self.dll.AUTDEmulatorSoundFieldInstant(record, range_, option)

    def emulator_sound_field_instant_time_len(self, sound_field: InstantPtr, duration: Duration) -> ctypes.c_uint64:
        return self.dll.AUTDEmulatorSoundFieldInstantTimeLen(sound_field, duration)

    def emulator_sound_field_instant_points_len(self, sound_field: InstantPtr) -> ctypes.c_uint64:
        return self.dll.AUTDEmulatorSoundFieldInstantPointsLen(sound_field)

    def emulator_sound_field_instant_get_x(self, sound_field: InstantPtr, x: ctypes.Array[ctypes.c_float]) -> None:
        return self.dll.AUTDEmulatorSoundFieldInstantGetX(sound_field, x)

    def emulator_sound_field_instant_get_y(self, sound_field: InstantPtr, y: ctypes.Array[ctypes.c_float]) -> None:
        return self.dll.AUTDEmulatorSoundFieldInstantGetY(sound_field, y)

    def emulator_sound_field_instant_get_z(self, sound_field: InstantPtr, z: ctypes.Array[ctypes.c_float]) -> None:
        return self.dll.AUTDEmulatorSoundFieldInstantGetZ(sound_field, z)

    def emulator_sound_field_instant_skip(self, sound_field: InstantPtr, duration: Duration) -> ResultStatus:
        return self.dll.AUTDEmulatorSoundFieldInstantSkip(sound_field, duration)

    def emulator_sound_field_instant_next(
        self,
        sound_field: InstantPtr,
        duration: Duration,
        time: ctypes.Array[ctypes.c_uint64],
        v: ctypes.Array[ctypes.Array[ctypes.c_float]],
    ) -> ResultStatus:
        return self.dll.AUTDEmulatorSoundFieldInstantNext(sound_field, duration, time, v)

    def emulator_sound_field_instant_free(self, sound_field: InstantPtr) -> None:
        return self.dll.AUTDEmulatorSoundFieldInstantFree(sound_field)

    def emulator_tracing_init(self) -> None:
        return self.dll.AUTDEmulatorTracingInit()

    def emulator_tracing_init_with_file(self, path: bytes) -> ResultStatus:
        return self.dll.AUTDEmulatorTracingInitWithFile(path)

    def emulator(self, pos: ctypes.Array[Point3], rot: ctypes.Array[Quaternion], len_: int) -> EmulatorPtr:
        return self.dll.AUTDEmulator(pos, rot, len_)

    def emulator_free(self, emulator: EmulatorPtr) -> None:
        return self.dll.AUTDEmulatorFree(emulator)

    def emulator_geometry(self, emulator: EmulatorPtr) -> GeometryPtr:
        return self.dll.AUTDEmulatorGeometry(emulator)

    def emulator_record_from(self, emulator: EmulatorPtr, start_time: DcSysTime, f: ctypes.c_void_p) -> ResultRecord:
        return self.dll.AUTDEmulatorRecordFrom(emulator, start_time, f)

    def emulator_record_free(self, record: RecordPtr) -> None:
        return self.dll.AUTDEmulatorRecordFree(record)

    def emulator_tick_ns(self, record: LinkPtr, tick: Duration) -> ResultStatus:
        return self.dll.AUTDEmulatorTickNs(record, tick)

    def emulator_transducer_table_rows(self, emulator: EmulatorPtr) -> ctypes.c_uint64:
        return self.dll.AUTDEmulatorTransducerTableRows(emulator)

    def emulator_transducer_table(
        self,
        emulator: EmulatorPtr,
        dev_indices: ctypes.Array[ctypes.c_uint16],
        tr_indices: ctypes.Array[ctypes.c_uint8],
        x: ctypes.Array[ctypes.c_float],
        y: ctypes.Array[ctypes.c_float],
        z: ctypes.Array[ctypes.c_float],
        nx: ctypes.Array[ctypes.c_float],
        ny: ctypes.Array[ctypes.c_float],
        nz: ctypes.Array[ctypes.c_float],
    ) -> None:
        return self.dll.AUTDEmulatorTransducerTable(emulator, dev_indices, tr_indices, x, y, z, nx, ny, nz)

    def emulator_record_output_cols(self, record: RecordPtr) -> ctypes.c_uint64:
        return self.dll.AUTDEmulatorRecordOutputCols(record)

    def emulator_record_output_voltage(self, record: RecordPtr, v: ctypes.Array[ctypes.Array[ctypes.c_float]]) -> None:
        return self.dll.AUTDEmulatorRecordOutputVoltage(record, v)

    def emulator_record_output_ultrasound(self, record: RecordPtr, v: ctypes.Array[ctypes.Array[ctypes.c_float]]) -> None:
        return self.dll.AUTDEmulatorRecordOutputUltrasound(record, v)

    def emulator_sound_field_rms(self, record: RecordPtr, range_: RangeXYZ, option: RmsRecordOption) -> ResultRms:
        return self.dll.AUTDEmulatorSoundFieldRms(record, range_, option)

    def emulator_sound_field_rms_time_len(self, sound_field: RmsPtr, duration: Duration) -> ctypes.c_uint64:
        return self.dll.AUTDEmulatorSoundFieldRmsTimeLen(sound_field, duration)

    def emulator_sound_field_rms_points_len(self, sound_field: RmsPtr) -> ctypes.c_uint64:
        return self.dll.AUTDEmulatorSoundFieldRmsPointsLen(sound_field)

    def emulator_sound_field_rms_get_x(self, sound_field: RmsPtr, x: ctypes.Array[ctypes.c_float]) -> None:
        return self.dll.AUTDEmulatorSoundFieldRmsGetX(sound_field, x)

    def emulator_sound_field_rms_get_y(self, sound_field: RmsPtr, y: ctypes.Array[ctypes.c_float]) -> None:
        return self.dll.AUTDEmulatorSoundFieldRmsGetY(sound_field, y)

    def emulator_sound_field_rms_get_z(self, sound_field: RmsPtr, z: ctypes.Array[ctypes.c_float]) -> None:
        return self.dll.AUTDEmulatorSoundFieldRmsGetZ(sound_field, z)

    def emulator_sound_field_rms_skip(self, sound_field: RmsPtr, duration: Duration) -> ResultStatus:
        return self.dll.AUTDEmulatorSoundFieldRmsSkip(sound_field, duration)

    def emulator_sound_field_rms_next(
        self,
        sound_field: RmsPtr,
        duration: Duration,
        time: ctypes.Array[ctypes.c_uint64],
        v: ctypes.Array[ctypes.Array[ctypes.c_float]],
    ) -> ResultStatus:
        return self.dll.AUTDEmulatorSoundFieldRmsNext(sound_field, duration, time, v)

    def emulator_sound_field_rms_free(self, sound_field: RmsPtr) -> None:
        return self.dll.AUTDEmulatorSoundFieldRmsFree(sound_field)



import ctypes
from typing import Self

import numpy as np
import polars as pl
from pyautd3.native_methods.utils import _validate_status
from pyautd3.utils import Duration

from pyautd3_emulator.native_methods.autd3capi_emulator import NativeMethods as Emu
from pyautd3_emulator.native_methods.autd3capi_emulator import RmsPtr
from pyautd3_emulator.native_methods.autd3capi_emulator import RmsRecordOption as RmsRecordOption_


class RmsRecordOption:
    _inner: RmsRecordOption_

    def __init__(
        self: Self,
        *,
        sound_speed: float = 340e3,
        print_progress: bool = False,
        gpu: bool = False,
    ) -> None:
        self._inner = RmsRecordOption_(
            sound_speed,
            print_progress,
            gpu,
        )


class Rms:
    _ptr: RmsPtr

    def __init__(self: Self, ptr: RmsPtr) -> None:
        self._ptr = ptr

    def __del__(self: Self) -> None:
        self._dispose()

    def _dispose(self: Self) -> None:
        if self._ptr.value is not None:  # pragma: no cover
            Emu().emulator_sound_field_rms_free(self._ptr)
            self._ptr.value = None

    def skip(self: Self, duration: Duration) -> Self:
        _validate_status(Emu().emulator_sound_field_rms_skip(self._ptr, duration._inner))
        return self

    def observe_points(self: Self) -> pl.DataFrame:
        points_len = int(Emu().emulator_sound_field_rms_points_len(self._ptr))
        x = np.zeros(points_len, dtype=np.float32)
        y = np.zeros(points_len, dtype=np.float32)
        z = np.zeros(points_len, dtype=np.float32)
        Emu().emulator_sound_field_rms_get_x(self._ptr, x.ctypes.data_as(ctypes.POINTER(ctypes.c_float)))  # type: ignore[arg-type]
        Emu().emulator_sound_field_rms_get_y(self._ptr, y.ctypes.data_as(ctypes.POINTER(ctypes.c_float)))  # type: ignore[arg-type]
        Emu().emulator_sound_field_rms_get_z(self._ptr, z.ctypes.data_as(ctypes.POINTER(ctypes.c_float)))  # type: ignore[arg-type]
        return pl.DataFrame(
            {
                "x[mm]": x,
                "y[mm]": y,
                "z[mm]": z,
            },
        )

    def next(self: Self, duration: Duration) -> pl.DataFrame:
        n = int(Emu().emulator_sound_field_rms_time_len(self._ptr, duration._inner))
        points_len = int(Emu().emulator_sound_field_rms_points_len(self._ptr))
        time = np.zeros(n, dtype=np.uint64)

        v = np.zeros([n, points_len], dtype=np.float32)
        _validate_status(
            Emu().emulator_sound_field_rms_next(
                self._ptr,
                duration._inner,
                time.ctypes.data_as(ctypes.POINTER(ctypes.c_uint64)),  # type: ignore[arg-type]
                ctypes.cast(
                    (ctypes.POINTER(ctypes.c_float) * n)(
                        *(ctypes.cast(r, ctypes.POINTER(ctypes.c_float)) for r in np.ctypeslib.as_ctypes(v)),  # type: ignore[arg-type]
                    ),
                    ctypes.POINTER(ctypes.POINTER(ctypes.c_float)),
                ),  # type: ignore[arg-type]
            ),
        )
        return pl.DataFrame(
            {s.name: s for s in (pl.Series(name=f"rms[Pa]@{time[i]}[ns]", values=v[i]) for i in range(n))},
        )



from typing import Self

from pyautd3 import Controller
from pyautd3.driver.link import Link
from pyautd3.native_methods.autd3capi_driver import ControllerPtr, GeometryPtr, LinkPtr
from pyautd3.native_methods.utils import _validate_status
from pyautd3.utils import Duration

from pyautd3_emulator.native_methods.autd3capi_emulator import NativeMethods as Emu


class RecorderLink(Link):
    def __init__(self: Self) -> None:
        super().__init__()

    def _resolve(self: Self) -> LinkPtr:
        err = "Not implemented"  # pragma: no cover
        raise NotImplementedError(err)  # pragma: no cover

    def tick(self: Self, tick: Duration) -> None:
        _validate_status(
            Emu().emulator_tick_ns(self._ptr, tick._inner),
        )


class Recorder(Controller[RecorderLink]):
    def __init__(self: Self, geometry: GeometryPtr, ptr: ControllerPtr) -> None:
        super().__init__(geometry, ptr, RecorderLink())

    def tick(self: Self, tick: Duration) -> None:
        self._link.tick(tick)



from typing import Self

from pyautd3_emulator.native_methods.autd3capi_emulator import RangeXYZ as RangeXYZ_


class RangeXYZ:
    _inner: RangeXYZ_

    def __init__(
        self: Self,
        *,
        x: tuple[float, float],
        y: tuple[float, float],
        z: tuple[float, float],
        resolution: float,
    ) -> None:
        self._inner = RangeXYZ_(
            x_start=x[0],
            x_end=x[1],
            y_start=y[0],
            y_end=y[1],
            z_start=z[0],
            z_end=z[1],
            resolution=resolution,
        )



import pyautd3

from pyautd3_emulator.native_methods.autd3capi_emulator import NativeMethods as Emu

from .emulator import Emulator
from .instant import InstantRecordOption
from .range import RangeXYZ
from .recorder import Recorder
from .rms import RmsRecordOption

pyautd3._ext_tracing_init.append(lambda: Emu().emulator_tracing_init())


__all__ = [
    "Emulator",
    "InstantRecordOption",
    "RangeXYZ",
    "Recorder",
    "RmsRecordOption",
]


__version__ = "30.0.1.2"
